package com.unitesk.requality.core;

import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.StringTokenizer;
import java.util.UUID;

import com.unitesk.requality.nodetypes.NodeType;
import com.unitesk.requality.nodetypes.Requirement;
import com.unitesk.requality.tools.DeepFirstTreeWalker;

public class TreeDB {
    private List<ITreeChangeListener> listeners = new ArrayList<ITreeChangeListener>();
    private TreeNode root;
    private Map<UUID, TreeNode> cache = new HashMap<UUID, TreeNode>();
    private TreeStorage storage;
    private Map<String, NodeType> defaultTypes = new HashMap<String, NodeType>();
    {
            NodeType dbroot = new NodeType("DBRoot", TreeDB.this);
            dbroot.setClass(com.unitesk.requality.core.TreeNode.class);
            defaultTypes.put("DBRoot", dbroot);
            NodeType nodetype = new NodeType("NodeType", TreeDB.this);
            nodetype.setClass(com.unitesk.requality.nodetypes.NodeType.class);
            defaultTypes.put("NodeType", nodetype);
            NodeType node = new NodeType("Node", TreeDB.this);
            node.setClass(com.unitesk.requality.core.TreeNode.class);
            defaultTypes.put("Node", node);
    };

    public boolean addListener(ITreeChangeListener listener) {
        return listeners.add(listener);
    }

    public boolean removeListener(ITreeChangeListener listener) {
        return listeners.remove(listener);
    }

    public List<ITreeChangeListener> getListeners() {
        return Collections.unmodifiableList(listeners);
    }

    public void addNodeType(String name, Class<? extends TreeNode> clazz) {
        NodeType nt = new NodeType(name, TreeDB.this);
        nt.setClass(clazz);
        defaultTypes.put(name, nt);
    }

    public void addNodeTypes(Map<String, Class<? extends TreeNode>> nts) {
        for(String name: nts.keySet()) {
            addNodeType(name, nts.get(name));
        }
    }

    private IStorageDeltaListener deltaProcessor = new IStorageDeltaListener() {
        @Override
        public void  created(UUID uuid) {
            String typeName = storage.getNodeType(uuid);
            NodeType nt;
            if (defaultTypes.containsKey(typeName)) {
                nt = defaultTypes.get(typeName);
            } else {
                nt = (NodeType) (getTypes().get(typeName));
            }
            TreeNode node = createFromStorage(uuid, nt);
            List<ITreeChangeListener> oldListeners = new ArrayList<ITreeChangeListener>(listeners);
            for (ITreeChangeListener l : oldListeners) {
                l.added(node);
            }
        }

        @Override
        public void deleted(UUID uuid) {
            TreeNode node = cache.get(uuid);
            node.getParent().children = new ArrayList<UUID>(
                    storage.getChildren(node.getParentUUId()));
            cache.remove(uuid);
            List<ITreeChangeListener> oldListeners = new ArrayList<ITreeChangeListener>(listeners);
            for (ITreeChangeListener l : oldListeners) {
                l.deleted(node);
            }
        }

        @Override
        public void changed(UUID uuid) {
            TreeNode node = cache.get(uuid);
            loadNodeData(uuid, node);
            List<ITreeChangeListener> oldListeners = new ArrayList<ITreeChangeListener>(listeners);
            for (ITreeChangeListener l : oldListeners) {
                l.changed(node);
            }
        }

        @Override
        public void moved(UUID uuid, UUID oldParentUUId, UUID newParent) {
            TreeNode node = cache.get(uuid);
            TreeNode oldParent = cache.get(oldParentUUId);
            node.parent = newParent;
            node.getParent().children = new ArrayList<UUID>(
                    storage.getChildren(newParent));
            oldParent.children = new ArrayList<UUID>(
                    storage.getChildren(oldParentUUId));

            List<ITreeChangeListener> oldListeners = new ArrayList<ITreeChangeListener>(listeners);
            for (ITreeChangeListener l : oldListeners) {
                l.moved(node, oldParent);
            }
        }
    };

    private void loadNodeData(UUID uuid, TreeNode node) {
        node.id = storage.getNodeId(uuid);
        for (Attribute a : storage.getAttributes(uuid)) {
            node.attributes.put(a.getKey(), a);
        }
    }

    public TreeNode createChildNode(UUID pUUId, String id, String type) {
        UUID uuid = UUID.randomUUID();
        if(getNode(pUUId).findChild(id)!=null) {
            throw new IllegalArgumentException("Child with id: "+" exists in "+getNode(pUUId).getQualifiedId());
        }
        storage.createNode(pUUId, uuid, type, id);
        return getNode(uuid);
    }

    public TreeDB(TreeStorage storage) {
        this.storage = storage;
        storage.setDeltaTarget(deltaProcessor);
    }
    public TreeNode getRootNode(boolean createIfNotExist) {
        if (root == null) {
            if (storage.getRootNode() == null && createIfNotExist) {
                UUID uuid = UUID.randomUUID();
                storage.createNode(null, uuid, "DBRoot", "/");
            }
            root = getNode(storage.getRootNode());
        }
        return root;
    }
    public TreeNode getRootNode() {
        return getRootNode(true);
    }

    public TreeNode getNode(UUID uuid) {
        if (cache.get(uuid) == null && storage.canLoad(uuid)) {
            String typeName = storage.getNodeType(uuid);
            NodeType nt;
            if (defaultTypes.containsKey(typeName)) {
                nt = defaultTypes.get(typeName);
            } else {
                nt = (NodeType) (getTypes().get(typeName));
            }
            createFromStorage(uuid, nt);
        }
        return cache.get(uuid);
    }

    public TreeNode getNode(String rootQId) {
        if (rootQId.equals("/")) {
            return getRootNode();
        }
        StringTokenizer st = new StringTokenizer(rootQId, "/");
        TreeNode cur = getRootNode();
        while (st.hasMoreTokens()) {
            String nodeId = st.nextToken();
            cur = cur.findChild(nodeId);
            if (cur == null) {
                return null;
            }
        }
        return cur;
    }

    private void cache(UUID uuid, TreeNode node) {
        cache.put(uuid, node);
    }

    public TreeStorage getStorage() {
        return storage;
    }

    public Map<String, NodeType> getTypes() {
        Map<String, NodeType> result = new HashMap<String, NodeType>(
                defaultTypes);
        TreeNode typesNode = getNode("/Types");
        if (typesNode != null && typesNode.sizeChildren() > 0) {
            for (int i = 0; i < typesNode.sizeChildren(); i++) {
                NodeType node = (NodeType) typesNode.getChild(i);
                result.put(node.getId(), node);
            }
        }
        return result;
    }

    public void createType(NodeType newType) {
        TreeNode types = getNode("/Types");
        if (types == null) {
            types = getRootNode().createChild("Types", "NodeType");
        }
        TreeNode newTypeNode = types.createChild(newType.getName());
        // TODO: (Types) store attribute constrains
        newTypeNode.saveAttributes();
    }

    public NodeType getNodeType(String name) {
        for (NodeType type : getTypes().values()) {
            if (type.getId().equals(name)) {
                return type;
            }
        }
        return null;
    }

    public boolean removeType(String typeName) {
        TreeNode types = getNode("/Types");
        if (types == null) {
            return false;
        }
        TreeNode n = types.findChild(typeName);
        n.delete();
        return true;
    }

    @SuppressWarnings("rawtypes")
    private TreeNode createFromStorage(UUID uuid, NodeType type) {
        // TODO: check out this code
        // if (cache.get(uuid) != null) {
        // throw new IllegalArgumentException("Node with uuid " + uuid
        // + " already exists");
        // }
        TreeNode result = null;
        try {
            Class[] treeNodeArgs = new Class[] { TreeDB.class, UUID.class };
            result = type.getTypeClass().getConstructor(treeNodeArgs)
                    .newInstance(TreeDB.this, uuid);
        } catch (Exception e) {
            e.printStackTrace();
        }
        if (result == null) {
            result = new TreeNode(this, uuid);
        }

        loadNodeData(uuid, result);

        result.parent = storage.getParent(uuid);
        result.children = new ArrayList<UUID>(storage.getChildren(uuid));
        if (result.getParent() != null) {
            result.getParent().children = new ArrayList<UUID>(
                    storage.getChildren(result.parent));
        }

        cache(uuid, result);
        if (result.getId().equals("/")) {
            // if (result.parent == null) {
            root = result;
        }
        return result;
    }

    // TODO: make static and move to Location
    public Requirement[] getRequirementsFor(String locQId) {
        ArrayList<Requirement> result = new ArrayList<Requirement>();
        for (TreeNode node : new DeepFirstTreeWalker(getNode(Requirement.getTypeRootQId()))) {
            if (!node.getType().equals("Requirement")) {
                continue;
            }
            Requirement req = (Requirement) node;
            if (req.getLocationQIds().contains(locQId)) {
                result.add(req);
            }
        }
        return result.toArray(new Requirement[0]);
    }

    public String getParentQualifiedId(String locQId) {
        return locQId.substring(0,locQId.lastIndexOf("/"));
    }
}
