/*
 * Copyright 2009 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package com.unitesk.requality.eclipse.editors;

import java.util.ArrayList;
import java.util.HashSet;
import java.util.List;
import java.util.Set;
import java.util.UUID;

import org.eclipse.core.runtime.CoreException;
import org.eclipse.core.runtime.IProgressMonitor;
import org.eclipse.swt.SWT;
import org.eclipse.swt.custom.ScrolledComposite;
import org.eclipse.swt.events.ControlAdapter;
import org.eclipse.swt.events.ControlEvent;
import org.eclipse.swt.graphics.Rectangle;
import org.eclipse.swt.layout.FillLayout;
import org.eclipse.swt.layout.GridData;
import org.eclipse.swt.layout.GridLayout;
import org.eclipse.swt.widgets.Composite;
import org.eclipse.swt.widgets.Display;
import org.eclipse.swt.widgets.Event;
import org.eclipse.swt.widgets.Listener;
import org.eclipse.ui.IEditorInput;
import org.eclipse.ui.IEditorSite;
import org.eclipse.ui.PartInitException;
import org.eclipse.ui.part.EditorPart;

import com.unitesk.requality.core.Attribute;
import com.unitesk.requality.core.AttributeType;
import com.unitesk.requality.core.ITreeChangeListener;
import com.unitesk.requality.core.TreeDB;
import com.unitesk.requality.core.TreeNode;
import com.unitesk.requality.eclipse.Activator;
import com.unitesk.requality.eclipse.editors.panels.BaseRequirementPanel;
import com.unitesk.requality.eclipse.editors.panels.RequirementAdditionalPropertiesPanel;
import com.unitesk.requality.eclipse.editors.panels.RequirementMainPropertiesPanel;
import com.unitesk.requality.eclipse.tools.DefaultTreeChangeListener;
import com.unitesk.requality.eclipse.tools.ITreeStateListener;
import com.unitesk.requality.eclipse.tools.TreesTracker;
import com.unitesk.requality.nodetypes.Requirement;

public class ReqEditor extends EditorPart {

    private Requirement req = null;
    private UUID reqUUId;

    private RequirementMainPropertiesPanel mainRequrementProperties;
    private RequirementAdditionalPropertiesPanel addinionalRequrementProperties;
    private BaseRequirementPanel baseRequirementProperties;

    private final ITreeStateListener treeStateListener = new ITreeStateListener() {
        @Override
        public void open(TreeDB db) {
        }

        @Override
        public void close(TreeDB db) {
            if (req.getTreeDB().equals(db)) {
                Display display = ReqEditor.this.getSite().getShell()
                        .getDisplay();
                if (!display.isDisposed()) {
                    display.syncExec(new Runnable() {
                        @Override
                        public void run() {
                            getSite().getPage().closeEditor(ReqEditor.this,
                                    false);
                        }
                    });
                }
            }
        }
    };
    private final ITreeChangeListener treeChangeListener = new DefaultTreeChangeListener() {
        @Override
        public void changed(TreeNode node) {
            if (req.getUUId().equals(node.getUUId())) {
                req = (Requirement) node;
                refresh();
            }
        }

        @Override
        public void deleted(TreeNode node) {
            if (reqUUId.equals(node.getUUId())) {
                Display display = ReqEditor.this.getSite().getShell()
                        .getDisplay();
                if (!display.isDisposed()) {
                    display.syncExec(new Runnable() {
                        @Override
                        public void run() {
                            getSite().getPage().closeEditor(ReqEditor.this,
                                    false);
                        }
                    });
                }
            }
        }
    };

    @Override
    public void doSave(IProgressMonitor monitor) {
        if (req != null && isChanged()) {
            Set<String> removedKeys = new HashSet<String>(
                    req.getAttributeKeys());
            removedKeys.removeAll(addinionalRequrementProperties
                    .getAttributes().keySet());
            for (String removeName : removedKeys) {
                if (removeName.startsWith("_")) {
                    continue;
                }
                req.removeAttribute(removeName);
            }
            for (String key : addinionalRequrementProperties.getAttributes()
                    .keySet()) {
                if (key.startsWith("_")) {
                    continue;
                }
                req.putAttribute(new Attribute(AttributeType.STRING, key,
                        addinionalRequrementProperties.getAttributes().get(key)));
            }

            if(!mainRequrementProperties.getReqName().equals(mainRequrementProperties.getId())) {
                req.setName(mainRequrementProperties.getReqName());
            } else {
                req.removeAttribute("_name");
            }
            req.setDescription(mainRequrementProperties.getDescription());
            req.setPredicate(addinionalRequrementProperties.getPredicate());

            List<String> toRemove = new ArrayList<String>(req.getBaseNodes());
            toRemove.removeAll(baseRequirementProperties.getBaseReqIds());

            List<String> toAdd = new ArrayList<String>(
                    baseRequirementProperties.getBaseReqIds());
            toAdd.removeAll(req.getBaseNodes());

            for (String removeId : toRemove)
                req.removeBaseNode(removeId);
            for (String addId : toAdd)
                req.addBaseNode(addId);

            req.saveAttributes();
            if (!req.getId().equals(mainRequrementProperties.getId())) {
                boolean needUpdateActiveNode = false;

                if (Activator.getDefault().getActiveNode("Requirement") == null) {
                    needUpdateActiveNode = true;
                } else if (Activator.getDefault().getActiveNode("Requirement")
                        .getQualifiedId().equals(req.getQualifiedId())) {
                    needUpdateActiveNode = true;
                }
                req.setId(mainRequrementProperties.getId());
                if (needUpdateActiveNode) {
                    Activator.getDefault().setActiveNode(req, "Requirement");
                }
            }
            setPartName(req.getName());
            firePropertyChange(PROP_DIRTY);
        }
    }

    @Override
    public void doSaveAs() {
        throw new UnsupportedOperationException();
    }

    @Override
    public void init(IEditorSite site, IEditorInput input)
            throws PartInitException {
        if (input instanceof NodeEditorInput) {
            NodeEditorInput nodeInput = (NodeEditorInput) input;
            try {
                setRequirement(nodeInput.getStorage().getNode());
            } catch (CoreException e) {
                e.printStackTrace();
            }
            setPartName(req.getName());
        } else
            throw new IllegalArgumentException(
                    "Input isn't instance of ReqEditorInput");
        setInput(input);
        setSite(site);

        req.getTreeDB().addListener(treeChangeListener);
        TreesTracker.getTracker().addListener(treeStateListener);
    }

    @Override
    public void dispose() {
        TreesTracker.getTracker().removeListener(treeStateListener);
        req.getTreeDB().removeListener(treeChangeListener);
        super.dispose();
    }
    private void setRequirement(Object req) {
        this.req = (Requirement) req;
        this.reqUUId = this.req.getUUId();
    }

    @Override
    public boolean isDirty() {
        return isChanged();
    }

    @Override
    public boolean isSaveAsAllowed() {
        return false;
    }

    @Override
    public void createPartControl(Composite parent) {
        parent.setLayout(new GridLayout());

        final ScrolledComposite scrollComposite = new ScrolledComposite(parent,
                SWT.V_SCROLL | SWT.H_SCROLL);
        scrollComposite.setLayout(new FillLayout());
        scrollComposite.setLayoutData(new GridData(GridData.FILL_BOTH));
        scrollComposite.setExpandHorizontal(true);
        scrollComposite.setExpandVertical(true);

        final Composite mainPanel = new Composite(scrollComposite, SWT.NONE);
        mainPanel.setLayout(new GridLayout());
        mainPanel.setLayoutData(new GridData(GridData.FILL_BOTH));

        Composite container = new Composite(mainPanel, SWT.NONE);
        container.setLayout(new GridLayout());
        container.setLayoutData(new GridData(GridData.FILL_BOTH));

        GridData mrpdata = new GridData(GridData.FILL_HORIZONTAL);
        mrpdata.minimumWidth = container.getSize().x;
        mainRequrementProperties = new RequirementMainPropertiesPanel(
                container, SWT.NONE);
        mainRequrementProperties.setLayoutData(mrpdata);
        mainRequrementProperties.setRequirement(req);
        addinionalRequrementProperties = new RequirementAdditionalPropertiesPanel(
                container, SWT.NONE);
        addinionalRequrementProperties.setLayoutData(new GridData(
                GridData.FILL_HORIZONTAL));

        addinionalRequrementProperties.setRequirement(req);

        addinionalRequrementProperties.addListener(SWT.Modify, modifyLink);

        GridData basedata = new GridData(GridData.FILL_HORIZONTAL);
        baseRequirementProperties = new BaseRequirementPanel(container,
                SWT.NONE);
        baseRequirementProperties.setLayoutData(basedata);
        baseRequirementProperties.setRequirement(req);

        mainRequrementProperties.addListener(SWT.Modify, modifyLink);

        baseRequirementProperties.addListener(SWT.Modify, modifyLink);

        mainPanel.pack();
        scrollComposite.setContent(mainPanel);
        scrollComposite.addControlListener(new ControlAdapter() {
            public void controlResized(ControlEvent e) {
                Rectangle r = scrollComposite.getClientArea();
                scrollComposite.setMinSize(mainPanel.computeSize(r.width,
                        SWT.DEFAULT));
            }
        });
    }

    @Override
    public void setFocus() {
        mainRequrementProperties.setFocus();
    }

    public void refresh() {
        // Если требование существует
        if (req != null && isChanged()==false) {

            if (mainRequrementProperties != null) {
                mainRequrementProperties.setRequirement(req);
                mainRequrementProperties.updateLocations(req);
            }
            if (addinionalRequrementProperties != null) {
                addinionalRequrementProperties.setRequirement(req);
            }
            if (baseRequirementProperties != null) {
                baseRequirementProperties.setRequirement(req);
            }
            firePropertyChange(PROP_DIRTY);
        }
    }

    protected boolean isChanged() {
        if (baseRequirementProperties == null
                || mainRequrementProperties == null
                || addinionalRequrementProperties == null) {
            return false;
        }

        if (baseRequirementProperties.isChanged())
            return true;
        if (mainRequrementProperties.isChanged())
            return true;
        if (addinionalRequrementProperties.isChanged())
            return true;
        return false;
    }

    private Listener modifyLink = new Listener() {
        public void handleEvent(Event event) {
            firePropertyChange(PROP_DIRTY);
        }
    };
}
